/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2001,2003 NoMachine, http://www.nomachine.com.           */
/*                                                                        */
/* NXPROXY, NX protocol compression and NX extensions to this software    */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for applicability.       */
/*                                                                        */
/* NX and NoMachine are trademarks of Medialogic S.p.A.                   */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/

#ifndef ServerChannel_H
#define ServerChannel_H

#include "Channel.h"

#include "SequenceQueue.h"

#include "ServerReadBuffer.h"

//
// Available unpack methods.
//

#include "Unpack.h"
#include "Tight.h"
#include "Jpeg.h"
#include "Png.h"

//
// Set the verbosity level.
//

#define PANIC
#undef  TEST
#undef  DEBUG

//
// How many sequence numbers of split commit
// requests we are going to save in order to
// mask errors.
//

#define MAX_COMMIT_SEQUENCE_QUEUE         16

//
// Define this to know when a channel
// is created or destroyed.
//

#undef  REFERENCES

//
// This class implements the X server
// side compression of X protocol.
//

class ServerChannel : public Channel
{
  public:

  ServerChannel(Transport *transport, Compressor *compressor,
                    Decompressor *decompressor);

  virtual ~ServerChannel();

  virtual int handleRead(EncodeBuffer &encodeBuffer);

  virtual int handleWrite(const unsigned char *message, unsigned int length);

  virtual int handleSplit(EncodeBuffer &encodeBuffer, MessageStore *store, 
                              const unsigned char *buffer, const unsigned int size);

  virtual int handleSplit(EncodeBuffer &encodeBuffer, int packetLimit);

  virtual int handleUnsplit(DecodeBuffer &decodeBuffer, MessageStore *store,
                                unsigned char *buffer, const unsigned int size);

  virtual int handleUnsplit(DecodeBuffer &decodeBuffer);

  virtual int needSplit() const
  {
    return (finish_ == 0 && clientStore_ ->
                getSplitStore() -> getAbort() > 0);
  }

  virtual int needMotion() const
  {
    return (finish_ == 0 && lastMotion_[0] != '\0');
  }

  virtual int needWakeup() const
  {
    return 0;
  }

  virtual int updateWakeup()
  {
    return 0;
  }

  virtual int needLimit()
  {
    return (transport_ -> length() >=
                control -> TransportXBufferLimit);
  }

  virtual int getLimit()
  {
    return (control -> TransportXBufferLimit);
  }

  virtual T_channel_type getType()
  {
    return CHANNEL_X;
  }

  void setBigEndian(int flag);

  protected:

  int handleFastReadReply(EncodeBuffer &encodeBuffer, const unsigned char &opcode,
                              const unsigned char *&buffer, const unsigned int &size);

  int handleFastReadEvent(EncodeBuffer &encodeBuffer, const unsigned char &opcode,
                              const unsigned char *&buffer, const unsigned int &size);

  int handleFastWriteRequest(DecodeBuffer &decodeBuffer, unsigned char &opcode,
                                 unsigned char *&buffer, unsigned int &size);

  //
  // Flush last motion notify event
  // to remote proxy.
  //

  int handleMotion(EncodeBuffer &encodeBuffer, int forceMotion = 0);

  virtual int handleWakeup(EncodeBuffer &encodeBuffer)
  {
    return 0;
  }

  virtual int handleConfiguration();

  virtual int handleFinish();

  virtual int handleNotify(T_notification_type type)
  {
    return 0;
  }
    
  virtual int handleReset();

  //
  // Synchronize last client's sequence
  // number with last request actually
  // received by X server.
  //

  int handleResetSequence(unsigned short int sequenceNum);

  //
  // Finalize the split operation after
  // the complete request has been
  // recomposed.
  //

  int handleCommit(DecodeBuffer &decodeBuffer, unsigned char &opcode,
                       unsigned char *&buffer, unsigned int &size);

  //
  // Uncompress a packed image in one
  // or more graphic X requests.
  //

  int handleUnpack(MessageStore *store, unsigned char *&buffer,
                       unsigned int &size);

  //
  // Model notification of congestion events.
  //

  virtual int isCongested()
  {
    int state = (transport_ -> blocked() == 1);

    if (state == 0 && control -> AgentSyncPropagate == 1)
    {
      if (syncCounter_ > 0 &&
              diffTimestamp(firstSyncTs_, getTimestamp()) >=
                  control -> CongestionTimeout)
      {
        #if defined(INFO) || defined(TEST)
        if (congestion_ == 0)
        {
          *logofs << "ServerChannel: Forcing channel for FD#" << fd_
                  << " in congestion with sync delay of "
                  << diffTimestamp(firstSyncTs_, getTimestamp())
                  << " Ms.\n" << logofs_flush;
        }
        #endif

        state = 1;
      }
    }

    return state;
  }

  virtual int isReliable()
  {
    return 1;
  }

  void handleEncodeCharInfo(const unsigned char *nextSrc, EncodeBuffer &encodeBuffer);

  //
  // Own read buffer. It is able to identify
  // full messages read from X descriptor.
  //

  ServerReadBuffer readBuffer_;

  //
  // Opcode of last request coming from X
  // client.
  //

  unsigned char clientOpcode_;
  unsigned char serverOpcode_;

  //
  // Sequence number of last request coming
  // from X client or X server.
  //

  unsigned int clientSequence_;
  unsigned int serverSequence_;

  //
  // Used to identify replies based on sequence
  // number of original request.
  //

  SequenceQueue sequenceQueue_;

  //
  // Last motion notify read from X server.
  //

  unsigned char lastMotion_[32];
  T_timestamp   lastMotionTs_;

  //
  // Sequence numbers of last auto-generated
  // put image requests. Needed to intercept
  // and suppress errors generated by such
  // requests.
  //

  unsigned int commitSequenceQueue_[MAX_COMMIT_SEQUENCE_QUEUE];

  //
  // Let agent select which expose
  // events is going to receive.
  //

  unsigned int enableExpose_;
  unsigned int enableGraphicsExpose_;
  unsigned int enableNoExpose_;

  //
  // Monitor sync replies coming from X server
  // in order to slow down the client peer in
  // case of congestion.
  //

  int syncCounter_;
  T_timestamp firstSyncTs_;

  //
  // Pixel geometry used to unpack
  // images.
  //

  T_geometry lastGeometry_;

  //
  // Current colormap, if any, used
  // to unpack images.
  //

  T_colormap lastColormap_;

  //
  // Unpack tight images.
  //

  Tight tightUnpacker_;

  private:

  //
  // Keep track of object creation and
  // deletion.
  //

  #ifdef REFERENCES

  static int references_;

  #endif
};

#endif /* ServerChannel_H */
